# STRUPHY - STRUcture-Preserving HYbrid codes

A Python package for energetic particles in plasma.

The package is developed since 2019 at [Max Planck Institute for Plasma Physics](https://www.ipp.mpg.de/) 
in the division [NMPP (Numerical Methods for Plasma Physics)](https://www.ipp.mpg.de/ippcms/de/for/bereiche/numerik).

## What you can do with Struphy

* Investigate plasma physics problems with multiple [models equations](https://struphy.pages.mpcdf.de/struphy/sections/models.html) containing fluid and/or kinetic components 
* Simulate each model on [curved domains](https://struphy.pages.mpcdf.de/struphy/sections/domains.html) 
* Load a variety of [MHD equilibira](https://struphy.pages.mpcdf.de/struphy/sections/mhd_equils.html#module-struphy.fields_background.mhd_equil.equils) and [kinetic backgrounds](https://struphy.pages.mpcdf.de/struphy/sections/kinetic_backgrounds.html)
* Use [diagnostic tools](https://struphy.pages.mpcdf.de/struphy/sections/diagnostics.html) to visualize results
* Compare simulations to analytic [dispersion relations](https://struphy.pages.mpcdf.de/struphy/sections/diagnostics.html#module-struphy.dispersion_relations.analytic)
* Solve the [MHD eigenvalue problem](https://struphy.pages.mpcdf.de/struphy/index.html) in axisymmetric toroidal geometries
* [Improve Struphy](https://struphy.pages.mpcdf.de/struphy/sections/developers.html) by adding
    * model equations
    * domains (geometry)
    * MHD equilibria
    * kinetic backgrounds
    * diagnostic tools
    * dispersion relations
    * other physics features (e.g. collision operators etc.)

## Algorithmic features

* Particle-in-cell method for kinetic species
* Discrete differential forms based on high-order B-spline finite elements ([Psydac library](https://github.com/pyccel/psydac)) for fields/fluids
* Exact conservation laws
* Polar splines to treat a polar singularity 
* Kernels are pre-compiled with [Pyccel](https://github.com/pyccel/pyccel) to achieve near-Fortran performance
* MPI/OpenMP hybrid parallelization  

## Installation

See the [Struphy documentation](https://struphy.pages.mpcdf.de/struphy/index.html).

## Tutorials

* [Tutorial 01](https://gitlab.mpcdf.mpg.de/struphy/struphy/-/blob/devel/notebooks/tutorial_01_units_run_main.ipynb): Struphy model's normalization (units), simulation parameters and the Struphy main execution file.

* [Tutorial 02](https://gitlab.mpcdf.mpg.de/struphy/struphy/-/blob/devel/notebooks/tutorial_02_postproc_standard_plotting.ipynb): Data generated by Struphy simulations, post processing and simple plotting of variables.

* [Tutorial 03](https://gitlab.mpcdf.mpg.de/struphy/struphy/-/blob/devel/notebooks/tutorial_03_plot_poloidal_planes.ipynb): Plotting field/fluid variables in toroidal geometry.

## Key references

* F. Holderied, S. Possanner, X. Wang, "MHD-kinetic hybrid code based on structure-preserving finite elements with particles-in-cell", [J. Comp. Phys. 433 (2021) 110143](https://www.sciencedirect.com/science/article/pii/S0021999121000358?via%3Dihub)

* F. Holderied, S. Possanner, "Magneto-hydrodynamic eigenvalue solver for axis-symmetric equilibria based on smooth polar splines", [J. Comp. Phys. 464 (2022) 111329](https://www.sciencedirect.com/science/article/pii/S0021999122003916?via%3Dihub)

* F. Holderied, "STRUPHY: a structure-preserving hybrid MHD-kinetic code for the interaction of energetic particles with Alfvén waves in magnetized plasmas", [PhD thesis (2022)](https://mediatum.ub.tum.de/?id=1656539)

## License

Copyright 2019 (c) Struphy dev team | Max Planck Institute for Plasma Physics

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

In addition, we ask you to cite the following reference in scientific publications which contain results obtained with
this software and developments:

F. Holderied, S. Possanner, X. Wang, "MHD-kinetic hybrid code based on structure-preserving finite elements with particles-in-cell", J. Comp. Phys. 433 (2021) 110143

## Contact

* Stefan Possanner [spossann@ipp.mpg.de](spossann@ipp.mpg.de)
* Eric Sonnendrücker [spossann@ipp.mpg.de](eric.sonnendruecker@ipp.mpg.de)
* Xin Wang [xin.wang@ipp.mpg.de](xin.wang@ipp.mpg.de)